/*
 * time.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/limits.h>
#include <sys/systm.h>
#include <isa/isareg.h>
#include <lib.h>
#include <errno.h>
#include <interrupt.h>
#include <proc.h>
#include <sp.h>
#include <mp.h>
#include <lock.h>
#include <mm.h>
#include <signal.h>
#include <debug.h>
#include <time.h>
#include <test.h>


//#define DEBUG_TIME 1

#ifdef DEBUG_TIME
	#define STATIC
	#define INLINE
#else
	#define STATIC static
	#define INLINE inline
#endif


STATIC uint setIntervalTimer(uint time);


/***************************************************************************************************
 *
 * ƥॿޡ
 *
 ***************************************************************************************************/


enum{
	TIMER_MAX = 10,			// ޡ
	REAL_TIMER_RATE	= 500,	/* Real-clock timer rate(ms) */

	/* Timer flags. */
	TIMER_PROC	= 1,	// ץޡץλ˺
	TIMER_ALONE,		// ɥ󥿥ޡץλʤ
	TIMER_DEL,			// 
};


typedef struct TIMER{
	struct TIMER *next;
	struct TIMER *prev;
	int flag;				// ޡ֥ե饰
	void *proc;				// ץ
	uint index;				// ޡֹ
	uint init;				// ޡ
	uint count;				// Ĥ
	struct TIMER *link;		// ³
	void (*handle)(void*);	// ϥɥ
	void *arg;				// ϥɥ
	int tasking;			// åʥ󥯸ξ YES = 롡NO = ʤ
}TIMER;


//================================== PRIVATE ============================================


static int timerCount;


// ƥॿޡƤ
// return : error number
STATIC int allocTimer(TIMER **o_timer)
{
	static uint index = 1;	// ޡֹ
	TIMER *timer;
	
	if (TIMER_MAX <= timerCount){
		// ɬƤ褦Ĵɬ
		WARNING(0);
		return -ENOMEM;
	}
	timer = kmalloc(sizeof(*timer));
	if (timer == NULL){
		return -ENOMEM;
	}
	memset(timer, 0, sizeof(*timer));
	timer->index = index++;
	++timerCount;
	
	*o_timer = timer;
/****************************************************************************************************************/
printDebug(80 * 1 + 70, "%d     ", timerCount);
/****************************************************************************************************************/
	
	return NOERR;
}


// ƥॿޡ
STATIC INLINE void freeTimer(TIMER *timer)
{
	kfree(timer);
	--timerCount;
	ASSERT(0 <= timerCount);
/****************************************************************************************************************/
printDebug(80 * 1 + 70, "%d     ", timerCount);
/****************************************************************************************************************/
}


//================================== PROTECTED ==========================================


static TIMER intervalTimer = {	// Interval timer link. */
	.next = &intervalTimer,
	.prev = &intervalTimer,
};
static TIMER realTimer = {		// Real timer link queue */
	.next = &realTimer,
	.prev = &realTimer,
};
static TIMER timeoutTimer = {	// ॢȤޡ
	.next = &timeoutTimer,
	.prev = &timeoutTimer,
};
static TIMER *linkTop[] = {
	&intervalTimer,
	&realTimer,
	&timeoutTimer,
};
static int timerLock;			// ޡԥå


//Գ
// 顼ϥɥ
STATIC void handleAlarm(void *proc)
{
	sendSignal(proc, SIGALRM);
}


//Գ
// ޡϥɥ
STATIC void handleWake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT | TASK_DEVICE_WAIT);
}

	
//Գ
// 󥰥ϥɥ
STATIC void handleTasking(
	void *i_link)		// ޡ󥯸
{
	TIMER *link = i_link;

	link->tasking = YES;
/*******************************************/
	{
		static char value = '0';

		if ('9' < value){
			value = '0';
		}
		displayTop(60, &value, 1);
		++value;
	}
/********************************************/
}


// ޡ³
//աtimerLock¾椵줿椫ƤӽФ뤳
STATIC void linkTimer(const uint i_time, TIMER *i_link, TIMER *m_timer)
{
	uint time = i_time;
	TIMER *nextTimer;
	
	ASSERT(i_link != NULL);
	ASSERT(m_timer != NULL);

	nextTimer = i_link->next;
	for (; nextTimer != i_link; nextTimer = nextTimer->next){
		if (time < nextTimer->count){
			nextTimer->count -= time;
			break;
		}
		else{
			time -= nextTimer->count;
		}
	}
	linkNextList(nextTimer, m_timer);		// Фꤹ
	m_timer->link = i_link;					// 󥯸
	m_timer->count = time;
}


// ޡڤΥ
//աtimerLock¾椵줿椫ƤӽФ뤳
STATIC void unlinkTimer(TIMER *m_timer)
{
	ASSERT(m_timer != NULL);

	m_timer->next->count += m_timer->count;
	unlinkDoublyList(m_timer);
}


// ޡץ򸵤õ
//աtimerLock¾椵줿椫ƤӽФ뤳
// return : error number
STATIC int searchTimerByProc(void (*handle)(), TIMER **o_timer)
{
	TIMER *timer;
	PROC *proc = get_current_task();
	int i;

	ASSERT(handle != NULL);

	for (i = 0; i < ARRAY_LENGTH(linkTop); ++i){
		for (timer = linkTop[i]->next; timer != linkTop[i] ; timer = timer->next){
			if ((timer->proc == proc) && (timer->handle == handle)){
				*o_timer = timer;
				return NOERR;
			}
		}
	}

	return -ENOENT;
}


// ޡֹõ
//աtimerLock¾椵줿椫ƤӽФ뤳
// return : error number
STATIC int searchTimerByIndex(const uint index, TIMER **o_timer)
{
	TIMER *timer;
	int i;

	for (i = 0; i < ARRAY_LENGTH(linkTop); ++i){
		for (timer = linkTop[i]->next; timer != linkTop[i] ; timer = timer->next){
			if (timer->index == index){
				*o_timer = timer;
				return NOERR;
			}
		}
	}

	return -ENOENT;
}


//ա
//  δؿɬ del_from_schedule() θǻȤȡ⤷˥ॢȤ
//  ץʤ롣
//  time0ʤʵפ˥꡼פ롣
// return : error number
STATIC int setTimer(const uint i_time, const uint initTime, const int flag, void (*handle)(), void *arg, TIMER **o_timer)
{
	TIMER  *link;
	TIMER  *timer;
	uint time = i_time;
	int eflag;
	int error;

	ASSERT(handle != NULL);
	ASSERT(TIMER_PROC <= flag && flag <= TIMER_DEL);

	if (time == 0){
		time = UINT_MAX;
	}

	// ³󥯸
	if (time < REAL_TIMER_RATE){
		link = &intervalTimer;
		time = setIntervalTimer(time);
	}
	else{
		link = &realTimer;
	}

	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		// ޡ
		if (flag == TIMER_PROC){
			// ³Ƥ륿ޡõ񤭤
			error = searchTimerByProc(handle, &timer);
			if (error == NOERR){
				unlinkTimer(timer);
				memset(timer, 0, sizeof(*timer));
			}
			else{
				error = allocTimer(&timer);
			}
		}
		else{
			error = allocTimer(&timer);
		}
		if (error != NOERR){
			// ɬƤ褦Ĵɬ
			WARNING(0);
			return error;
		}

		// 󥯤³
		linkTimer(time, link, timer);
		timer->flag = flag;
		timer->init = initTime;
		timer->handle = handle;
		timer->arg = arg;
		timer->proc = get_current_task();
    }
	exit_spinlock(&timerLock);
	exitCli(eflag);

	*o_timer = timer;

	return NOERR;
}


// Ĥ֤
STATIC void getTimerRest(TIMER *timer, struct itimerval *value)
{
	uint sec, mili, time = 0;
	int eflag;
	
	ASSERT(timer != NULL);
	ASSERT(value != NULL);

	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		if (timer->next != timer){
			TIMER *p;

			// Ĥ֤û
			for (p = timer; p != timer->link; p = p->prev){
				time += p->count;
			}
		}
	}
	exit_spinlock(&timerLock);
	exitCli(eflag);

	sec = timer->init / 1000;
	mili = timer->init - sec * 1000;
	value->it_interval.tv_sec = sec;
	value->it_interval.tv_usec = mili * 1000;

	sec = time / 1000;
	mili = time - sec * 1000;
	value->it_value.tv_sec = sec;
	value->it_value.tv_usec = mili * 1000;
}


STATIC INLINE void delTimer(TIMER *timer)
{
	ASSERT(timer != NULL);
	timer->flag = TIMER_DEL;
}


// ॢȤǺѤߤΥޡ
/*STATIC void releaseTimeoutTimer()
{
	TIMER *timer;
	TIMER *save;
	int eflag;
	
	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		for (timer = timeoutTimer.next; timer != &timeoutTimer ;){
			save = timer->next;
			if (timer->flag == TIMER_DEL){
				unlinkDoublyList(timer);
				freeTimer(timer);
			}
			timer = save;
		}
	}
	exit_spinlock(&timerLock);
	exitCli(eflag);
}
*/

// ֤вᤷޡϥɥ¹Ԥ
//աtimerLock¾椵줿椫ƤӽФ뤳
//Գ
STATIC void execTimer(
	TIMER		*link,		// 󥯸
	const uint	elapseTime)	// в
{
	static struct handler{
		void (*handle)(void*);
		void *arg;
	}exec[TIMER_MAX];
	int c_exec = 0;
	TIMER *linkWait = NULL;
	TIMER *timer;
	TIMER *save;
	uint elapse = elapseTime;
	int eflag;
	int i;

	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		for (timer = link->next; timer != link;){
			if (timer->count <= elapse){
				save = timer->next;
				unlinkDoublyList(timer);
				elapse -= timer->count;
				timer->count = 0;
				if (timer->flag != TIMER_DEL){
					if (c_exec < TIMER_MAX){
						exec[c_exec].handle = timer->handle;
						exec[c_exec].arg = timer->arg;
						++c_exec;
					}
					else{
						// ޡƤƤΤ̵ʤ뤳ȤϤʤϤ
						ASSERT(0);
					}

					// ͤȤԤ󥯤ˤĤʤǸǥåȤ
					if (timer->init != 0){
						timer->next = linkWait;
						linkWait = timer;
					}
					else{
						if (timer->flag == TIMER_ALONE){
							freeTimer(timer);
						}
						else{
							// ॢȥ󥯤³
							linkPrevList(&timeoutTimer, timer);	// Фꤹ
							timer->link = &timeoutTimer;		// 󥯸
						}
					}
				}
				else{
					freeTimer(timer);
				}

				timer = save;
			}
			else{
				timer->count -= elapse;
				break;
			}
		}

		// ͤΤ󥯤ľ
		for (timer = linkWait; timer != NULL; timer = save){
			save = timer->next;
			linkTimer(timer->init, link, timer);
		}
	}
	exit_spinlock(&timerLock);
	exitCli(eflag);

	// ޡϥɥ¹Ԥ
	for (i = 0; i < c_exec; ++i){
		exec[i].handle(exec[i].arg);
	}
}


// ƥॿޡγ
STATIC void releaseSysTimer()
{
	PROC *proc = get_current_task();
	TIMER *timer;
	int eflag;
	int i;

	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		for (i = 0; i < ARRAY_LENGTH(linkTop); ++i){
			for (timer = linkTop[i]->next; timer != linkTop[i] ;){
				 TIMER *save = timer->next;
				if ((timer->proc == proc) && (timer->flag == TIMER_PROC)){
					unlinkTimer(timer);
					freeTimer(timer);
				}
				timer = save;
			}
		}
    }
	exit_spinlock(&timerLock);
	exitCli(eflag);
}


//================================== PUBLIC =============================================


// 󥰤
// return : error number
int setTasking(
	const uint	i_time)		// 󥰻֡ms
{
	TIMER *timer;
	int error;
	
	// ޡκ
	error = setTimer(i_time, i_time, TIMER_ALONE, handleTasking, &intervalTimer, &timer);
	if (error != NOERR){
		return error;
	}

	return NOERR;
}


// Wake timer from device waiting.
// error number
int setWakeTimer(
	const uint	i_time,		// ޡ֡ms
	void		**o_timer)	// ޡ
{
	TIMER *timer;
	int error;
	
	// ޡκ
	error = setTimer(i_time, 0, TIMER_PROC, handleWake, get_current_task(), &timer);
	if (error != NOERR){
		return error;
	}
	*o_timer = timer;

	return NOERR;
}


// Delete timer.
void delWakeTimer(
	void *i_timer)		// ͭʥޡ
{
	delTimer(i_timer);
}


// Execute a function after a specified length of time.
//Գ
//ա¾ץƤФ礢
struct callout_handle timeout(timeout_t *ftn, void *arg, int to_ticks)
{
	TIMER *timer;
	struct callout_handle callout;

	if (setTimer(to_ticks, 0, TIMER_ALONE, ftn, arg, &timer) == NOERR){
		callout.callout = timer;
		callout.index = timer->index;
	}
	else{
		printk("Failed timeout()!\n");
		callout.callout = NULL;
	}

	return callout;
}


//Գ
// Cancel previous timeout function call.
void untimeout(timeout_t *ftn, void *arg, struct callout_handle callout)
{
	TIMER *timer;
	int eflag;

	if (callout.callout == NULL){
		return;
	}
	eflag = enterCli();
	enter_spinlock(&timerLock);
	{
		if (searchTimerByIndex(callout.index, &timer) == NOERR){
			unlinkTimer(timer);
			freeTimer(timer);
		}
    }
	exit_spinlock(&timerLock);
	exitCli(eflag);
}


//	Initialize a handle so that using it with untimeout is benign.
void callout_handle_init(struct callout_handle *callout)
{
	callout->callout = NULL;
}


/***************************************************************************************************
 *
 * 8254󥿡Х륿ޡ
 *
 ***************************************************************************************************/


//================================== PRIVATE ============================================


static int intervalTime;		// ߤ֡ms


// ֡msˤ򥯥åѴ롣
// return : å
STATIC INLINE uint timeToClock(
	const uint i_time)		// ms
{
	return PIT_HZ * i_time / 1000;
}


// å֡msˤѴ롣
// return : ֡ms
STATIC INLINE uint clockToTime(
	const uint i_clock)		// å
{
	return i_clock * 1000 / PIT_HZ;
}


// 8254 interval timer handler
// return : YES = å, NO = åʤ
STATIC int interval_timer_handler()
{
	intervalTimer.tasking = NO;

	execTimer(&intervalTimer, intervalTime);

	if (intervalTimer.next != &intervalTimer){
		uint clock;

		// ޡꤹ
		intervalTime = intervalTimer.next->count;
		clock = timeToClock(intervalTime);
		outb(PIT_CNR0, clock & 0xff);
		outb(PIT_CNR0, clock >> 8);
	}
	else{
		intervalTime = 0;
	}

	return intervalTimer.tasking;
}


//================================== PROTECTED ==========================================


// Set interval timer.
// return : Ĵᤷ֡ms
STATIC uint setIntervalTimer(
	const uint i_time)		// ֡ms
{
	int restClock;
	int clock = timeToClock(i_time);

	if (intervalTime == 0){
		intervalTime = i_time;
		outb(PIT_CNR0, clock & 0xff);
		outb(PIT_CNR0, clock >> 8);
		return intervalTime;
	}

	restClock = inb(PIT_CNR0);		// Ĥꥯå
	restClock += inb(PIT_CNR0) << 8;

	if (clock < restClock){
		intervalTime -= clockToTime(restClock - clock);
		if (0 < intervalTime){
			outb(PIT_CNR0, clock & 0xff);
			outb(PIT_CNR0, clock >> 8);
		}
		else{
			intervalTime += clockToTime(restClock - clock);
		}
		return intervalTime;
	}
	else{
		return intervalTime + clockToTime(clock - restClock);
	}
}


//================================== PUBLIC =============================================


// 
void initIntervalTimer()
{
	irq_entry[IRQ0] = interval_timer_handler;
	outb(PIT_CTR,0x30);		// One shot timer(mode 0).
//	outb(PIT_CTR,0x34);		// CLK0,LSB.MSB,⡼2,binary.
	release_irq_mask(0);
}


/***************************************************************************************************
 *
 * MC146818ꥢ륯åޡ
 *
 ***************************************************************************************************/


//================================== PRIVATE ============================================


static uint utc;				/* current UTC(number of seconds since 00:00:00.1.1.1970). */
static uint64 curRdtsc;			/* ߤRDTSC͡ */


/*
 * ʿʿѴ롣
 */
STATIC INLINE uchar bcd_to_bin(uchar value)
{
	return value - 6 * (value >> 4);
}


// MC146818 real timer handler
// return : ץåʤ
STATIC int real_timer_handler()
{
	static int half = 0;
	uint64 half64;

	// Status register C ɤǳߤꥻåȤ롣
	read_cmos(CMOS_STRC);

	// ޡϥɥμ¹
	execTimer(&realTimer, REAL_TIMER_RATE);

	// ॢȥޡ
//	releaseTimeoutTimer();

	/* UTC򹹿롣 */
	half ^= 1;		/* 500msȹ */
	utc += half;
	
	/* ߤRDTSCͤ */
	half64 = half;
	curRdtsc *= half64 ^ 1;
	curRdtsc += rdtsc() * half64;
	
	return 0;
}


//================================== PROTECTED ==========================================


/*
 * UTCꤹ롣
 * parameters : seconds
 */
STATIC void INLINE setUtc(uint seconds)
{
	utc = seconds;
}


/*
 * get current time from cmos
 * parameters : time structure
 */
STATIC void get_time_from_cmos()
{
	/* ι */
	static ushort sumMonth[2][13]=
	{{0,0,31,59,90,120,151,181,212,243,273,304,334},
	 {0,0,31,60,91,121,152,182,213,244,274,305,335},
	};

	struct TIME{
		uchar second;		/* seconds */
		uchar minute;		/* minutes */
		uchar hour;			/* hours */
		uchar day;			/* days */
		uchar month;		/* months */
		ushort year;		/* years */
	}time;
	int days;

	do{
		time.second = bcd_to_bin(read_cmos(CMOS_SCD));
		time.minute = bcd_to_bin(read_cmos(CMOS_MNT));
		time.hour = bcd_to_bin(read_cmos(CMOS_HOR));
		time.day = bcd_to_bin(read_cmos(CMOS_DAY));
		time.month = bcd_to_bin(read_cmos(CMOS_MTH));
		time.year = bcd_to_bin(read_cmos(CMOS_YER));
	}while (bcd_to_bin(read_cmos(CMOS_SCD)) < time.second);

	time.year += (time.year < 70)? 2000 : 1900;

	/* UTCη׻ */
	days = (time.year - 1970) * 365;
	days += (time.year - 1 - 1968) / 4;
	days += sumMonth[(time.year % 4) == 0][time.month];
	days += time.day - 1;
	utc = ((days * 24 + time.hour) * 60 + time.minute) * 60 + time.second;

	/* GMTĴ */
	utc -= LOCAL_TIME_TOKYO;
}


//================================== PUBLIC =============================================


/*
 * CMOS read and write
 */
uchar read_cmos(int addr)
{
	outb(CMOS_ADR, addr | NMI_ENABLE);
	return inb(CMOS_DTR);
}

void write_cmos(int addr, uchar value)
{
	outb(CMOS_ADR, addr | NMI_ENABLE);
	outb(CMOS_DTR, value);
}


void gettimeofday(struct timeval *m_tp)
{
	m_tp->tv_sec = utc;
	m_tp->tv_usec = (rdtsc() - curRdtsc) / clock_1micro;
	if (1000000 <= m_tp->tv_usec){
		m_tp->tv_sec += 1;
		m_tp->tv_usec -= 1000000;
	}
}


void getmicrouptime(struct timeval *m_tp)
{
	uint64 current = rdtsc();
	
	m_tp->tv_sec = current / (clock_1micro * 1000000);
	m_tp->tv_usec = (current % (clock_1micro * 1000000)) / clock_1micro;
}


int rtcin(int reg)
{
	u_char val;

	outb(IO_RTC, reg);
	inb(0x84);
	val = inb(IO_RTC + 1);
	inb(0x84);

	return (val);
}


void initRealTimer()
{
	irq_entry[IRQ8] = real_timer_handler;
	get_time_from_cmos();				/* get current time from cmos. */
	write_cmos(CMOS_STRA, 0x2f);		/* time bese 0x010(default)|ޡ졼65536/(20xf)=2HZ(500ms) */
	read_cmos(CMOS_STRC);				/* Status register C ɤǳߤꥻåȤ롣 */
	write_cmos(CMOS_STRB, 0x42);		/* 24hour,ߥ */
	release_irq_mask(8);
}


/***************************************************************************************************
 *
 * ƥ।󥿡Х륿ޡ
 *
 ***************************************************************************************************/


//================================== PRIVATE ============================================


static ITIMER itimerReal[MAX_CPU];		// ꥢ륤󥿡Х륿ޡ󥯸
static int itimerRock[MAX_CPU];			// ꥢ륤󥿡Х륿ޡѥå
static uint64 beforeRealRdtsc[MAX_CPU];		// ľΥꥢ륤󥿡Х륿ޡrdtsc


// 󥿡Х륿ޡ¹Ԥ
STATIC void doProcItimer(
	const int64 time,	// в֡rdtsc͡
	const int signal,	// ޡв륷ʥ
	ITIMER *m_itimer)
{
	if (m_itimer->cur <= 0){
		return;
	}
	m_itimer->cur -= time;
	if (m_itimer->cur <= 0){
		sendSignal(m_itimer->proc, signal);
		m_itimer->cur = m_itimer->init;
	}
}


//================================== PROTECTED ==========================================


STATIC void getItimer(int which, struct itimerval *m_value)
{
	time_t sec;
	suseconds_t micro;
	ITIMER *itimer = &get_current_task()->itimer[which];
	
	micro = itimer->init / clock_1micro;
	sec = micro / 1000000;
	m_value->it_interval.tv_sec = sec;
	m_value->it_interval.tv_usec = micro % 1000000;

	micro = itimer->cur / clock_1micro;
	sec = micro / 1000000;
	m_value->it_value.tv_sec = sec;
	m_value->it_value.tv_usec = micro % 1000000;
}


STATIC void setItimer(const int which, const struct itimerval *value, struct itimerval *m_value)
{
	uint64 micro;
	ITIMER *itimer = &get_current_task()->itimer[which];
	int cpu = get_current_cpu();
	
	/* ޡͤ */
	if (m_value != NULL){
		getItimer(which, m_value);
	}
	
	/* ޡͤ */
	micro = value->it_interval.tv_sec * 1000000 + value->it_interval.tv_usec;
	itimer->init = micro * clock_1micro;
	micro = value->it_value.tv_sec * 1000000 + value->it_value.tv_usec;
	itimer->cur = micro * clock_1micro;

	// ꥢ륿ޡ󥯤³
	if (which == ITIMER_REAL){
		int eflag = enterCli();
		enter_spinlock(&itimerRock[cpu]);
		{
			if (0 < itimer->cur){
				linkNextList(&itimerReal[get_current_cpu()], itimer);
			}
			else{
				unlinkDoublyList(itimer);
			}
		}
		exit_spinlock(&itimerRock[cpu]);
		exitCli(eflag);
	}
}


// ץ󥿡Х륿ޡν
STATIC void initProcItimer(PROC *proc)
{
	ITIMER *itimer = proc->itimer;

	memset(itimer, 0, sizeof(proc->itimer));
	initlinkList(&itimer[ITIMER_REAL]);
	itimer[ITIMER_REAL].proc = proc;
}


// ץ󥿡Х륿ޡγ
STATIC void releaseProcItimer(ITIMER *itimer)
{
	int cpu = get_current_cpu();
	int eflag;
	
	eflag = enterCli();
	enter_spinlock(&itimerRock[cpu]);
	{
		unlinkDoublyList(&itimer[ITIMER_REAL]);
	}
	exit_spinlock(&itimerRock[cpu]);
	exitCli(eflag);
}


// 󥿡Х륿ޡν
STATIC void initItimer()
{
	int i;

	for (i = 0; i < MAX_CPU; i++){
		initlinkList(&itimerReal[i]);
	}
}


//================================== PUBLIC =============================================


// ꥢ륤󥿡Х륿ޡ¹Ԥ
void doRealItimer()
{
	int cpu = get_current_cpu();
	const int64 curTime = rdtsc();
	const int64 time = curTime - beforeRealRdtsc[cpu];
	int eflag;
	ITIMER *itimer;
	
	eflag = enterCli();
	enter_spinlock(&itimerRock[cpu]);
	{
		for (itimer = itimerReal[cpu].next; itimer != &itimerReal[cpu];){
			ITIMER *next = itimer->next;

			itimer->cur -= time;
			if (itimer->cur <= 0){
				sendSignal(itimer->proc, SIGALRM);
				itimer->cur = itimer->init;
			}
			if (itimer->cur <= 0){
				unlinkDoublyList(itimer);
			}
			itimer = next;
		}
	}
	exit_spinlock(&itimerRock[cpu]);
	exitCli(eflag);
	
	beforeRealRdtsc[cpu] = curTime;
}


// ƥв֤ˤ륤󥿡Х륿ޡ
void doSystemItimer(
	int64 time,
	ITIMER *itimer)		// в֡rdtsc͡
{
	// ꥢ륤󥿡Х륿ޡμ¹
	doProcItimer(time, SIGPROF, &itimer[ITIMER_PROF]);
}


// 桼в֤ˤ륤󥿡Х륿ޡ
void doUserItimer(
	int64 time,
	ITIMER *itimer)		// в֡rdtsc͡
{
	// ꥢ륤󥿡Х륿ޡμ¹
	doProcItimer(time, SIGVTALRM, &itimer[ITIMER_VIRTUAL]);
	doProcItimer(time, SIGPROF, &itimer[ITIMER_PROF]);
}


/***************************************************************************************************
 *
 * 
 *
 ***************************************************************************************************/


//================================== PROTECTED ==========================================


//FreeBSD
STATIC void timevalfix(struct timeval *t1)
{
	if (t1->tv_usec < 0) {
		t1->tv_sec--;
		t1->tv_usec += 1000000;
	}
	if (t1->tv_usec >= 1000000) {
		t1->tv_sec++;
		t1->tv_usec -= 1000000;
	}
}


//================================== PUBLIC =============================================


/*
 * Add and subtract routines for timevals.
 * N.B.: subtract routine doesn't deal with
 * results which are before the beginning,
 * it just gets very confused in this case.
 * Caveat emptor.
 */
//FreeBSD
void timevaladd(struct timeval *t1, struct timeval *t2)
{
	t1->tv_sec += t2->tv_sec;
	t1->tv_usec += t2->tv_usec;
	timevalfix(t1);
}


//FreeBSD
void timevalsub(struct timeval *t1, struct timeval *t2)
{
	t1->tv_sec -= t2->tv_sec;
  	t1->tv_usec -= t2->tv_usec;
	timevalfix(t1);
}


/***************************************************************************************************
 *
 * å¬
 *
 ***************************************************************************************************/


//================================== PUBLIC =============================================


int clock_1micro;	/* 1micros */
int clock_1m;		/* 1ms */


/*
 * cpuå¬
 * ֤ : å
 */
int count_cpu_clock()
{
	uint cpu_clock;
	int low,high;
	long long r_beg,r_end;
	double time;

 	outb(PIT_CTR,0x30);	/* CLK0,LSB.MSB,⡼0,binary */
	outb(PIT_CNR0,0);	/* LSB */
	outb(PIT_CNR0,0);	/* MSB */

 	r_beg = rdtsc();	/* clock counter */

	for (;;){
		if (0x400000 < rdtsc() - r_beg){	/* ٱ */
			break;
		}
	}

	r_end = rdtsc();

	low=inb(PIT_CNR0);
	high=inb(PIT_CNR0);

	time=(0xffff-low-(high<<8))/(double)PIT_HZ;
	cpu_clock=(r_end-r_beg)/time;

 	clock_1micro = cpu_clock / 1000000 + 1;		/* 1micros */
 	clock_1m = cpu_clock / 1000 + 1;			/* 1ms */

	return cpu_clock;
}


/***************************************************************************************************
 *
 * ץ
 *
 ***************************************************************************************************/


// ץޡطν
void initProcTimer(void *i_proc)
{
	initProcItimer(i_proc);
}


// ץޡطγ
void releaseProcTimer()
{
	releaseSysTimer();
	releaseProcItimer(get_current_task()->itimer);
}


/***************************************************************************************************
 *
 * ƥॳ
 *
 ***************************************************************************************************/


int sys_time(int *tp)
{
	if(tp != NULL)
	{
		if (checkMem(tp,sizeof(int)) == -1)
			return -EFAULT;
		*tp = utc;
	}

	return (int)utc;
}


int sys_sleep(uint seconds)
{
	TIMER *timer;
	struct itimerval value;
	int time;

	if (setTimer(seconds * 1000, 0, TIMER_PROC, handleWake, get_current_task(), &timer) != NOERR){
		return seconds;
	}

	sigWait();

	// Ĥ֤
	getTimerRest(timer, &value);
	delTimer(timer);
	time = value.it_value.tv_sec;

	return time;
}


//Գ
void sys_wake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT);
}


int sys_alarm(uint seconds)
{
	TIMER *timer;
	int error;

	if (0 < seconds){
		setTimer(seconds*1000, 0, TIMER_PROC, handleAlarm, get_current_task(), &timer);
	}
	else{
		int eflag;
		
		eflag = enterCli();
		enter_spinlock(&timerLock);
		{
			 error = searchTimerByProc(handleAlarm, &timer);
	    }
		exit_spinlock(&timerLock);
		exitCli(eflag);
		if (error == NOERR){
			delTimer(timer);
		}
	}
	
	return 0;
}


int sys_settime(uint seconds)
{
	if (get_current_task()->uid!=0){
		return -EPERM;
	}
	setUtc(seconds);

	return 0;
}


int sys_gettimeofday(struct timeval *tp, void *tzp)
{
	if (checkMem(tp, sizeof(struct timeval)) == -1){
		return 0;
	}
	gettimeofday(tp);

	return 0;	
}


int sys_getitimer(int which, struct itimerval *m_value)
{
	if (checkMem(m_value, sizeof(*m_value)) == -1){
		return -EFAULT;
	}
	if (ITIMER_PROF < which){
		return -EFAULT;
	}

	getItimer(which, m_value);

	return 0;
}


int sys_setitimer(int which, struct itimerval *value, struct itimerval *m_value)
{
	if (checkMem(value, sizeof(*value)) == -1){
		return -EFAULT;
	}
	if (m_value != NULL){
		if (checkMem(m_value, sizeof(*m_value)) == -1){
			return -EFAULT;
		}
	}
	if (ITIMER_PROF < which){
		return -EFAULT;
	}

	setItimer(which, value, m_value);

	return 0;
}


/***************************************************************************************************
 *
 * 
 *
 ***************************************************************************************************/


void initTimer()
{
	initItimer();
}


/*************************************************************************************/
void test_time(uint signal)
{
	if(get_current_task()==(PROC*)0x150400)
	{
		printk("dosignal signal=%x ",signal);
	}	
}
/**************************************************************************************/
