/*
 * Copyright (c) 1987, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)disklabel.h	8.2 (Berkeley) 7/10/94
 * $FreeBSD: src/sys/sys/disklabel.h,v 1.49.2.7 2001/05/27 05:58:26 jkh Exp $
 */


#ifndef _SYS_DISKLABEL_H_
#define	_SYS_DISKLABEL_H_


#include <sys/cdefs.h>
#include <sys/types.h>


#ifdef KERNEL


/* XXX these should be defined per controller (or drive) elsewhere, not here! */
#define LABELSECTOR	1			/* sector containing label */
#define LABELOFFSET	0			/* offset of label in sector */
#define DISKMAGIC	((u_int32_t)0x82564557)	/* The disk magic number */
#define	MAXPARTITIONS	8

/* d_type values: */
#define	DTYPE_SMD		1		/* SMD, XSMD; VAX hp/up */
#define	DTYPE_MSCP		2		/* MSCP */
#define	DTYPE_DEC		3		/* other DEC (rk, rl) */
#define	DTYPE_SCSI		4		/* SCSI */
#define	DTYPE_ESDI		5		/* ESDI interface */
#define	DTYPE_ST506		6		/* ST506 etc. */
#define	DTYPE_HPIB		7		/* CS/80 on HP-IB */
#define	DTYPE_HPFL		8		/* HP Fiber-link */
#define	DTYPE_FLOPPY	10		/* floppy */
#define	DTYPE_CCD		11		/* concatenated disk */
#define	DTYPE_VINUM		12		/* vinum volume */
#define	DTYPE_DOC2K		13		/* Msys DiskOnChip */

/*
 * Filesystem type and version.
 * Used to interpret other filesystem-specific
 * per-partition information.
 */
#define	FS_UNUSED	0		/* unused */
#define	FS_SWAP		1		/* swap */
#define	FS_V6		2		/* Sixth Edition */
#define	FS_V7		3		/* Seventh Edition */
#define	FS_SYSV		4		/* System V */
#define	FS_V71K		5		/* V7 with 1K blocks (4.1, 2.9) */
#define	FS_V8		6		/* Eighth Edition, 4K blocks */
#define	FS_BSDFFS	7		/* 4.2BSD fast file system */
#define	FS_MSDOS	8		/* MSDOS file system */
#define	FS_BSDLFS	9		/* 4.4BSD log-structured file system */
#define	FS_OTHER	10		/* in use, but unknown/unsupported */
#define	FS_HPFS		11		/* OS/2 high-performance file system */
#define	FS_ISO9660	12		/* ISO 9660, normally CD-ROM */
#define	FS_BOOT		13		/* partition contains bootstrap */
#define	FS_VINUM	14		/* Vinum drive */

/*
 * flags shared by various drives:
 */
#define		D_REMOVABLE	0x01		/* removable media */
#define		D_ECC		0x02		/* supports ECC */
#define		D_BADSECT	0x04		/* supports bad sector forw. */
#define		D_RAMDISK	0x08		/* disk emulator */
#define		D_CHAIN		0x10		/* can do back-back transfers */

#define	LABEL_PART	2		/* partition containing label */
#define	RAW_PART	2		/* partition containing whole disk */
#define	SWAP_PART	1		/* partition normally containing swap */

#define DIOCGDINFO		_IOR('d', 101, struct disklabel)/* get */
#define DIOCSDINFO		_IOW('d', 102, struct disklabel)/* set */
#define DIOCWDINFO		_IOW('d', 103, struct disklabel)/* set, update disk */
#define DIOCGPART		_IOW('d', 104, struct partinfo)	/* get partition */
#define DIOCGDVIRGIN	_IOR('d', 105, struct disklabel) /* get virgin label */
#define DIOCWLABEL		_IOW('d', 109, int)	/* write en/disable label */


struct disklabel {
	u_int32_t	d_magic;		/* the magic number */
	u_int16_t	d_type;			/* drive type */
	u_int16_t	d_subtype;		/* controller/d_type specific */
	char		d_typename[16];	/* type name, e.g. "eagle" */
	/*
	 * d_packname contains the pack identifier and is returned when
	 * the disklabel is read off the disk or in-core copy.
	 * d_boot0 and d_boot1 are the (optional) names of the
	 * primary (block 0) and secondary (block 1-15) bootstraps
	 * as found in /boot.  These are returned when using
	 * getdiskbyname(3) to retrieve the values from /etc/disktab.
	 */
	union {
		char	un_d_packname[16];	/* pack identifier */
		struct {
			char	*un_d_boot0;	/* primary bootstrap name */
			char	*un_d_boot1;	/* secondary bootstrap name */
		} un_b;
	} d_un;
#define d_packname	d_un.un_d_packname
#define d_boot0		d_un.un_b.un_d_boot0
#define d_boot1		d_un.un_b.un_d_boot1
		/* disk geometry: */
	u_int32_t	d_secsize;		/* # of bytes per sector */
	u_int32_t	d_nsectors;		/* # of data sectors per track */
	u_int32_t	d_ntracks;		/* # of tracks per cylinder */
	u_int32_t	d_ncylinders;	/* # of data cylinders per unit */
	u_int32_t	d_secpercyl;	/* # of data sectors per cylinder */
	u_int32_t	d_secperunit;	/* # of data sectors per unit */
	/*
	 * Spares (bad sector replacements) below are not counted in
	 * d_nsectors or d_secpercyl.  Spare sectors are assumed to
	 * be physical sectors which occupy space at the end of each
	 * track and/or cylinder.
	 */
	u_int16_t	d_sparespertrack;	/* # of spare sectors per track */
	u_int16_t	d_sparespercyl;		/* # of spare sectors per cylinder */
	/*
	 * Alternate cylinders include maintenance, replacement, configuration
	 * description areas, etc.
	 */
	u_int32_t	d_acylinders;		/* # of alt. cylinders per unit */

		/* hardware characteristics: */
	/*
	 * d_interleave, d_trackskew and d_cylskew describe perturbations
	 * in the media format used to compensate for a slow controller.
	 * Interleave is physical sector interleave, set up by the
	 * formatter or controller when formatting.  When interleaving is
	 * in use, logically adjacent sectors are not physically
	 * contiguous, but instead are separated by some number of
	 * sectors.  It is specified as the ratio of physical sectors
	 * traversed per logical sector.  Thus an interleave of 1:1
	 * implies contiguous layout, while 2:1 implies that logical
	 * sector 0 is separated by one sector from logical sector 1.
	 * d_trackskew is the offset of sector 0 on track N relative to
	 * sector 0 on track N-1 on the same cylinder.  Finally, d_cylskew
	 * is the offset of sector 0 on cylinder N relative to sector 0
	 * on cylinder N-1.
	 */
	u_int16_t	d_rpm;				/* rotational speed */
	u_int16_t	d_interleave;		/* hardware sector interleave */
	u_int16_t	d_trackskew;		/* sector 0 skew, per track */
	u_int16_t	d_cylskew;			/* sector 0 skew, per cylinder */
	u_int32_t	d_headswitch;		/* head switch time, usec */
	u_int32_t	d_trkseek;			/* track-to-track seek, usec */
	u_int32_t	d_flags;			/* generic flags */
#define NDDATA 5
	u_int32_t	d_drivedata[NDDATA];	/* drive-type specific information */
#define NSPARE 5
	u_int32_t	d_spare[NSPARE];	/* reserved for future use */
	u_int32_t	d_magic2;			/* the magic number (again) */
	u_int16_t	d_checksum;			/* xor of data incl. partitions */

			/* filesystem and partition information: */
	u_int16_t	d_npartitions;		/* number of partitions in following */
	u_int32_t	d_bbsize;			/* size of boot area at sn0, bytes */
	u_int32_t	d_sbsize;			/* max size of fs superblock, bytes */
	struct	partition {				/* the partition table */
		u_int32_t	p_size;			/* number of sectors in partition */
		u_int32_t	p_offset;		/* starting sector */
		u_int32_t	p_fsize;		/* filesystem basic fragment size */
		u_int8_t	p_fstype;		/* filesystem type, see below */
		u_int8_t	p_frag;			/* filesystem fragments per block */
		union {
			u_int16_t	cpg;		/* UFS: FS cylinders per group */
			u_int16_t	sgs;		/* LFS: FS segment shift */
		} __partition_u1;
#define	p_cpg	__partition_u1.cpg
#define	p_sgs	__partition_u1.sgs
	} d_partitions[MAXPARTITIONS];	/* actually may be more */
};


static inline u_int16_t dkcksum(struct disklabel *lp)
{
	u_int16_t *start = (u_int16_t *)lp;
	u_int16_t *end = (u_int16_t *)&lp->d_partitions[lp->d_npartitions];
	u_int16_t sum = 0;

	while (start < end){
		sum ^= *start++;
	}

	return (sum);
}

#define	dkpart(dev)		(minor(dev) & 7)
#define	dkslice(dev)	((minor(dev) >> 16) & 0x1f)

static inline u_int dkunit(dev_t dev)
{
	return (((minor(dev) >> 16) & 0x1e0) | ((minor(dev) >> 3) & 0x1f));
}


extern void	bufqdisksort __P((struct buf_queue_head *ap, struct buf *bp));
extern char	*readdisklabel __P((dev_t dev, struct disklabel *lp));
extern int	setdisklabel __P((struct disklabel *olp, struct disklabel *nlp, u_long openmask));
extern int	writedisklabel __P((dev_t dev, struct disklabel *lp));
extern void	diskerr __P((struct buf *bp, char *what, int pri, int blkdone, struct disklabel *lp));


#endif /* !KERNEL */


#endif /* !_SYS_DISKLABEL_H_ */
